"""This script enables cancellation of all currently running skills."""
from typing import Callable


def cancel_skills(misty: Callable):
    """Cancels all skills currently running on Misty."""
    data = misty.get_info("skills_running")
    result = data.get("result", [])
    to_cancel = []
    for dct in result:
        uid = dct.get("uniqueId", "")
        if len(uid) > 0:
            to_cancel.append(uid)
    for skill in to_cancel:
        misty.perform_action("skill_cancel", data={"Skill": skill})



"""The main script for the misty2py package.
"""
from misty2py.utils.messages import compose_json_reply, message_parser
from misty2py.utils.generators import get_random_string
from typing import Callable
from misty2py.action import *
from misty2py.information import *
from misty2py.misty_event import MistyEventHandler


class Misty:
    """A class representing a Misty robot.

    Attributes:
        ip (str): The IP address of this Misty robot.
        infos (Info()): The object of Info class that belongs to this Misty.
        actions (Action()): The object of Action class that belongs
        to this Misty.
        events (dict): A dictionary of active event subscriptions (keys being
        the event name, values the MistyEvent() object).
    """

    def __init__(
        self,
        ip: str,
        custom_info: Dict = {},
        custom_actions: Dict = {},
        custom_data: Dict = {},
    ):
        """Initialises an instance of a Misty robot.

        Args:
            ip (str): The IP address of this Misty robot.
            custom_info (Dict, optional): Custom information keywords
            in the form of dictionary with key being the keyword and value
            being the API endpoint. Defaults to {}.
            custom_actions (Dict, optional): Custom actions keywords
            in the form of dictionary with key being the keyword and value
            being the API endpoint. Defaults to {}.
            custom_data (Dict, optional): Custom data shortcuts in the form
            of dictionary with key being the shortcut and value being the json
            data in the form of a dictionary. Defaults to {}.
        """

        self.ip = ip
        self.infos = Info(ip, custom_allowed_infos=custom_info)
        self.actions = Action(
            ip,
            custom_allowed_actions=custom_actions,
            custom_allowed_data=custom_data
        )
        self.event_handler = MistyEventHandler(ip)

    def __str__(self) -> str:
        """Transforms a Misty() object into a string.

        Returns:
            str: A string identifiyng this Misty robot
        """

        return "A Misty II robot with IP address %s" % self.ip

    def perform_action(self, action_name: str, data: Dict = {}) -> Dict:
        """Sends Misty a request to perform an action.

        Args:
            action_name (str): The keyword specifying the action to perform.
            data (Dict, optional): The data to send in the request body
            in the form of a data shortcut or a json dictionary.
            Defaults to {}.

        Returns:
            Dict: response from the API
        """

        return self.actions.action_handler(action_name, data)

    def get_info(self, info_name: str, params: Dict = {}) -> Dict:
        """Obtains information from Misty.

        Args:
            info_name (str): The information keyword specifying which kind
            of information to retrieve.
            params (Dict): dict of parameter name and parameter value.
            Defaults to {}.

        Returns:
            Dict: The requested information in the form of a json dictionary.
        """

        return self.infos.get_info(info_name, params)

    def event(self, action: str, **kwargs) -> Dict:
        """Handles event-related actions.

        Supports following actions:
            - event subscripton: requires an action keyword "subscribe"
            and an argument "type" (a string representing the event type
            to subscribe to). Optional arguments are:
                - name (str) for a custom event name; must be unique.
                - return_property (str) for the property to return from
                Misty's websockets; all properties are returned if
                return_property is not supplied.
                - debounce (int) for the interval at which new information is
                sent in ms; defaults to 250.
                - len_data_entries (int) for the maximum number of data
                entries to keep (discards in fifo style); defaults to 10.
                - event_emitter (Callable) for an event emitter function which
                emits an event upon message recieval.
            - obtaining the data from an event: requires an action keyword
            "get_data" and an argument "name" (the event name).
            - obtaining the log from an event: requires an action keyword
            "get_log" and an argument "name" (the event name).
            - unsubscribing from an event: requires an action keyword
            "unsubscribe" and an argument "name" (the event name).

        Args:
            action (str): the action keyword.

        Returns:
            Dict: the json dict response.
        """

        if action == "subscribe":
            return self.event_handler.subscribe_event(kwargs)

        if action == "get_data":
            return self.event_handler.get_event_data(kwargs)

        if action == "get_log":
            return self.event_handler.get_event_log(kwargs)

        if action == "unsubscribe":
            return self.event_handler.unsubscribe_event(kwargs)

        return {
            "status": "Failed",
            "message": "Unknown event action: `%s`." % action
        }



"""base64 encoders and decoders"""
import base64
from typing import Union
from os import listdir
from os.path import isfile, join


def content_to_base64(
    input_str: str, is_path: bool = False, save_path: Union[str, bool] = False
) -> str:
    """Encodes content to base64.

    Args:
        input_str (str): the utf-8 string to encode or the path to the file
        to encode.
        is_path (bool, optional): True if input_str is the path to the file
        to encode, False if input_str is the content to encode. Defaults
        to False.
        save_path (Union[str,bool], optional): path to the file where
        the encoded content should be saved, False if the encoded content
        should be returned as utf-8 string. Defaults to False.

    Returns:
        str: either the encoded content if save_path is False, or the success
        message if save_path is True.
    """
    if is_path:
        try:
            data = open(input_str, "rb").read()
        except Exception as e:
            return "Error reading the file `%s`. Error message: `%s`" % (
                input_str,
                e
            )
    else:
        data = input_str.encode()
    encoded = base64.b64encode(data)
    if save_path:
        try:
            with open(save_path, "wb") as f:
                f.write(encoded)
            return "Successfully saved to `%s`" % save_path
        except Exception as e:
            return "Error saving to `%s`. Error message: `%s`" % (
                save_path,
                e
            )
    return encoded.decode("utf-8")


def base64_to_content(
    input_str: str, is_path: bool = False, save_path: Union[str, bool] = False
) -> str:
    """Decodes base64 to utf-8 string or other content type specified
    by the extension in save_path.

    Args:
        input_str (str): the utf-8 string to decode or the path to the file
        to decode.
        is_path (bool, optional): True if input_str is the path to the file
        to decode, False if input_str is the content to decode. Defaults
        to False.
        save_path (Union[str,bool], optional): path to the file where
        the decoded content should be saved, False if the decoded content
        should be returned as utf-8 string. Defaults to False.

    Returns:
        str: [description]
    """
    if is_path:
        try:
            data = open(input_str, "rb").read()
        except Exception as e:
            return "Error reading the file `%s`. Error message: `%s`" % (
                input_str,
                e
            )
    else:
        data = input_str.encode()
    decoded = base64.b64decode(data)
    if save_path:
        try:
            with open(save_path, "wb") as f:
                f.write(decoded)
            return "Successfully saved to `%s`" % save_path
        except Exception as e:
            return "Error saving to `%s`. Error message: `%s`" % (
                save_path,
                e
            )
    return decoded.decode("utf-8")



"""Utility functions that do not fit anywhere else."""
import os
from typing import Any, Callable, Dict, List, Type

from dotenv import dotenv_values


def get_project_folder(env_path: str = ".env") -> str:
    """Obtains the project directory usinf the PROJECT_DIR value
    from the supplied .env file."""
    values = dotenv_values(env_path)
    potential_path = values.get("PROJECT_DIR", "./")
    if os.path.isdir(potential_path):
        return os.path.abspath(potential_path)
    return os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def get_abs_path(rel_path: str) -> str:
    """Constructs the absolute path from a relative path."""
    return os.path.abspath(os.path.join(get_project_folder(), rel_path))


def get_misty(env_path: str = ".env") -> Callable:
    """Obtains a Misty instance using the MISTY_IP_ADDRESS in the supplied
    .env file."""
    from misty2py.robot import Misty
    from misty2py.utils.env_loader import EnvLoader

    env_loader = EnvLoader(get_abs_path(env_path))
    return Misty(env_loader.get_ip())


def get_files_in_dir(abs_dir: str) -> List[str]:
    """Lists files in the supplied directory specified with absolute path."""
    return [
        os.path.join(abs_dir, f)
        for f in os.listdir(abs_dir)
        if os.path.isfile(os.path.join(abs_dir, f))
    ]


def get_base_fname_without_ext(fname: str) -> str:
    """Returns the file name without the extension (or extensions if multiple)
    from the supplied path."""
    base = os.path.basename(fname)
    return os.path.splitext(base)[0]


def query_dict_with_fallback(
    data: Dict, query: str, fallback: Any, required_type: Type = None
) -> Any:
    """Safely queries a dictionary, returns the queried value if it passes
    type check and the fallback otherwise.

    Args:
        data (Dict): the dictionary to query.
        query (str): the query.
        fallback (Any): the fallback value to return if the key does not exist
        or the queried value does not pass the type check.
        required_type (Type, optional): The required type of the queried
        value. None if there is no required type. Defaults to None.

    Returns:
        Any: the query result.
    """
    result = data.get(query)

    if result is None:
        return fallback

    if required_type is not None:
        if not isinstance(result, required_type):
            return fallback

    return result



"""EnvLoader class"""
from typing import Optional

from dotenv import dotenv_values


class EnvLoader:
    """Loads environmental values.

    Attributes:
        values [OrderedDict]: The loaded environmental values, might be empty.
    """

    def __init__(self, env_path: str = ".env") -> None:
        self.values = dotenv_values(env_path)

    def get_ip(self) -> Optional[str]:
        """Obtains MISTY_IP_ADDRESS from environmental values."""
        return self.values.get("MISTY_IP_ADDRESS")



"""The script that enables speech with extended return message.
"""
from typing import Callable, Dict

from misty2py.utils.generators import get_random_string
from misty2py.utils.messages import compose_json_reply


def speak(misty: Callable, utterance: str) -> Dict:
    """Speaks an utterance and returns descriptive message.

    Args:
        utterance (str): The utterance to speak.

    Returns:
        Dict: the response, including the "status" keyword and a "message"
        keyword which contains the utterance spoken.
    """
    result = misty.perform_action(
        "speak",
        data={
            "Text": utterance,
            "UtteranceId": "utterance_" + get_random_string(6)
            },
    )
    return compose_json_reply(
        result,
        success_message="Talking successful. Utterance: `%s`." % utterance,
        fail_message="Talking failed. Utterance: `%s`." % utterance,
    )



"""This script enables the websocket information recieval from Misty.
"""
import json
import threading
from typing import Callable, Dict, Union

import websocket

from misty2py.utils.generators import get_random_string


class MistyEvent:
    """A class that represents an event type subscribed to.

    Attributes:
        server (str): Misty's websocket server address.
        data (list): The data entries obtained.
        type_str (str): The event type string as required by Misty's
        websockets.
        event_name (str): A custom, unique event name.
        return_property (str): The property to return as requeired by Misty's
        websockets.
        debounce (int): The interval at which new information is sent in ms.
        log (list): The logs.
        len_data_entries (int): The maximum number of data entries to keep.
        ee (Union[bool, Callable]): The event emitter function if one is
        desired, False otherwise.
    """

    def __init__(
        self,
        ip: str,
        type_str: str,
        event_name: str,
        return_property: str,
        debounce: int,
        len_data_entries: int,
        event_emitter: Union[Callable, None],
    ):
        """Initialises an event object.

        Args:
            ip (str): Misty's IP address.
            type_str (str): The event type string as required by Misty's
            websockets.
            event_name (str): A custom, unique event name.
            return_property (str): The property to return as required by
            Misty's websockets.
            debounce (int): The interval at which new information is sent
            in ms.
            len_data_entries (int): The maximum number of data entries to keep.
            event_emitter (Union[Callable, None]): The event emitter function
            if one is desired, False otherwise.
        """

        self.server = "ws://%s/pubsub" % ip
        self.data = []
        self.type_str = type_str
        self.event_name = event_name
        self.return_property = return_property
        self.debounce = debounce
        self.log = []
        self.len_data_entries = len_data_entries
        event_thread = threading.Thread(target=self.run, daemon=True)
        event_thread.start()
        if event_emitter:
            self.ee = event_emitter
        else:
            self.ee = False

    def run(self):
        """Initialises the subscription and data collection."""

        self.ws = websocket.WebSocketApp(
            self.server,
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
        )
        self.ws.run_forever()

    def on_message(self, ws, message):
        """Saves received data and if ee is set, emits an event.

        Args:
            ws (websocket.WebSocketApp): the event's websocket.
            message (str): the data.
        """

        message = json.loads(message)
        mes = message["message"]
        if len(self.data) > self.len_data_entries:
            self.data = self.data[1:-1]
        self.data.append(mes)

        if self.ee:
            self.ee.emit(self.event_name, mes)

    def on_error(self, ws, error):
        """Logs an error and if ee is set, emits an 'error' event.

        Args:
            ws (websocket.WebSocketApp): the event type's websocket.
            error (str): the error message.
        """

        if len(self.log) > self.len_data_entries:
            self.log = self.log[1:-1]
        self.log.append(error)

        if self.ee:
            self.ee.emit("error_%s" % self.event_name, error)

    def on_close(self, ws):
        """Appends the closing message to the log and if ee is set,
        emits a 'close' event.

        Args:
            ws (websocket.WebSocketApp): the event type's websocket.
        """

        mes = "Closed"
        if len(self.log) > self.len_data_entries:
            self.log = self.log[1:-1]
        self.log.append(mes)

        if self.ee:
            self.ee.emit("close_%s" % self.event_name, mes)

    def on_open(self, ws):
        """Appends the opening message to the log and starts the subscription
        and if ee is set, emits an 'open' event.

        Args:
            ws (websocket.WebSocketApp): the event type's websocket.
        """

        self.log.append("Opened")
        self.subscribe()
        ws.send("")

        if self.ee:
            self.ee.emit("open_%s" % self.event_name)

    def subscribe(self):
        """Constructs the subscription message."""

        msg = {
            "Operation": "subscribe",
            "Type": self.type_str,
            "DebounceMs": self.debounce,
            "EventName": self.event_name,
            "ReturnProperty": self.return_property,
        }
        msg_str = json.dumps(msg, separators=(",", ":"))
        self.ws.send(msg_str)

    def unsubscribe(self):
        """Constructs the unsubscription message."""

        msg = {
            "Operation": "unsubscribe",
            "EventName": self.event_name,
            "Message": ""
        }
        msg_str = json.dumps(msg, separators=(",", ":"))
        self.ws.send(msg_str)
        self.ws.close()


class MistyEventHandler:
    """A class that handles all events its related Misty object
    subscribed to during this runtime.

    Attributes:
        events (Dict): a dictionary of all Event objects its related Misty
        object subscribed to during this runtime.
        ip (str): the IP address of the Misty belonging to this object's
        related Misty object.
    """

    def __init__(self, ip: str):
        """Initialises an object of class MistyEventHandler.

        Args:
            ip (str): the IP address of the Misty belonging to this object's
            related Misty object.
        """

        self.events = {}
        self.ip = ip

    def subscribe_event(self, kwargs: Dict) -> Dict:
        """Subscribes to an event type.

        Args:
            kwargs (Dict):  requires a key "type" (a string representing the
            event type to subscribe to). Optional keys are:
                - name (str) for a custom event name; must be unique.
                - return_property (str) for the property to return from
                Misty's websockets; all properties are returned if
                return_property is not supplied.
                - debounce (int) for the interval at which new information is
                sent in ms; defaults to 250.
                - len_data_entries (int) for the maximum number of data
                entries to keep (discards in fifo style); defaults to 10.
                - event_emitter (Callable) for an event emitter function which
                emits an event upon message recieval.

        Returns:
            Dict: success/fail message in the form of json dict.
        """

        event_type = kwargs.get("type")
        if not event_type:
            return {"status": "Failed", "message": "No event type specified."}

        event_name = kwargs.get("name")
        if not event_name:
            event_name = "event_%s_%s" % (event_type, get_random_string(8))

        return_property = kwargs.get("return_property")

        debounce = kwargs.get("debounce")
        if not debounce:
            debounce = 250

        len_data_entries = kwargs.get("len_data_entries")
        if not len_data_entries:
            len_data_entries = 10

        event_emitter = kwargs.get("event_emitter")

        try:
            new_event = MistyEvent(
                self.ip,
                event_type,
                event_name,
                return_property,
                debounce,
                len_data_entries,
                event_emitter,
            )

        except Exception as e:
            return {
                "status": "Failed",
                "message": "Error occurred while attempting \
                    to subscribe to an event of type `%s`. \
                        Error message: `%s`"
                % (
                    event_type,
                    e
                ),
            }

        self.events[event_name] = new_event

        return {
            "status": "Success",
            "message": "Subscribed to event type `%s` with name `%s`"
            % (event_type, event_name),
            "event_name": event_name,
        }

    def get_event_data(self, kwargs: Dict) -> Dict:
        """Obtains data from a subscribed event type.

        Args:
            kwargs (Dict): Requires a key "name" (the event name).

        Returns:
            Dict: the data or the fail message in the form of a json dict.
        """

        event_name = kwargs.get("name")
        if not event_name:
            return {"status": "Failed", "message": "No event name specified."}

        if event_name in self.events.keys():
            try:
                return {
                    "status": "Success",
                    "message": self.events[event_name].data
                }
            except Exception as e:
                return {
                    "status": "Failed",
                    "message": "Error occurred. Error message: `%s`" % e
                }

        else:
            return {
                "status": "Failed",
                "message": "Event type `%s` is not subscribed to."
                % event_name,
            }

    def get_event_log(self, kwargs: Dict) -> Dict:
        """Obtains the log from a subscribed event type.

        Args:
            kwargs (Dict): Requires a key "name" (the event name).

        Returns:
            Dict: the log or the fail message in the form of a json dict.
        """

        event_name = kwargs.get("name")
        if not event_name:
            return {"status": "Failed", "message": "No event name specified."}

        if event_name in self.events.keys():
            try:
                return {
                    "status": "Success",
                    "message": self.events[event_name].log
                }
            except Exception as e:
                return {
                    "status": "Failed",
                    "message": "Error occurred while attempting \
                        to access the log of event `%s`. \
                            Error message: `%s`"
                    % (
                        event_name,
                        e
                    ),
                }

        else:
            return {
                "status": "Failed",
                "message": "Event `%s` is not subscribed to." % event_name,
            }

    def unsubscribe_event(self, kwargs: Dict) -> Dict:
        """Unsubscribes from an event type.

        Args:
            kwargs (Dict): Requires a key "name" (the event name).

        Returns:
            Dict: success/fail message in the form of json dict.
        """

        event_name = kwargs.get("name")
        if not event_name:
            return {"status": "Failed", "message": "No event name specified."}

        if event_name in self.events.keys():
            try:
                self.events[event_name].unsubscribe()
                mes = {
                    "status": "Success",
                    "message": "Event `%s` of type `%s` unsubscribed"
                    % (event_name, self.events[event_name].type_str),
                    "log": self.events[event_name].log,
                }
            except Exception as e:
                mes = {
                    "status": "Failed",
                    "message": "Error occurred while attempting \
                        to unsubscribe from event `%s`. \
                            Error message: `%s`"
                    % (
                        event_name,
                        e
                    ),
                }
            self.events.pop(event_name)
            return mes
        else:
            return {
                "status": "Failed",
                "message": "Event type `%s` is not subscribed to."
                % event_name,
            }



"""Removes non-system files from Misty's memory."""
import os
from typing import Callable, Dict, List, Tuple

from misty2py.utils.messages import success_parser_from_dicts
from misty2py.utils.utils import get_abs_path


def get_non_system_assets(
    misty: Callable,
    assets: List = ["audio", "image", "video", "recording"],
) -> Dict:
    """Obtains the list of all non-system assets that fall under one of
    the `assets` types.

    Args:
        misty (Callable): An instance of misty2py.robot's Misty class.
        assets (List, optional): The types of assets to obtain.
        Defaults to ["audio", "image", "video", "recording"].

    Returns:
        Dict: all non-system assets of a given type (keywords) listed
        as their names in the list under the type keyword.
    """
    action = "_list"
    non_sys_assets = {}
    for asset in assets:
        data = misty.get_info(asset + action)
        result = data.get("result", [])
        for hit in result:
            name = hit.get("name")
            if not name:
                continue

            is_system = hit.get("systemAsset", True)
            if not is_system:
                if not non_sys_assets.get(asset):
                    non_sys_assets[asset] = []
                non_sys_assets[asset].append(name)
                print(
                    "Found a non-system asset \
                        of type `%s` named `%s`." % (asset, name)
                )

    return non_sys_assets


def get_asset_properties(asset_type: str, file: str) -> Tuple[Dict, str, str]:
    """Obtains the parameters, the name and the extension of a file.

    Args:
        asset_type (str): The type of the asset.
        file (str): The file name.

    Returns:
        Tuple[Dict, str, str]: the parameters (a dict with keys "Name"
        for recordings or "FileName" for other asset types containing the name
        and the key "Base64" set to "true"), the file name and the file
        extension.
    """
    if asset_type == "recording":
        params = {"Name": file, "Base64": "true"}
    else:
        params = {"FileName": file, "Base64": "true"}

    split_file_name = file.split(".")
    name = split_file_name[0]
    if len(split_file_name) > 1:
        ext = split_file_name[1]
    else:
        ext = "unknown"

    return params, name, ext


def save_base64_str(
    full_path: str,
    content: str,
    overwrite: bool = False
) -> bool:
    """Saves an asset content to the file specified by `full_path`.

    Args:
        full_path (str): The absolute path to the save location.
        content (str): The base64 content to save.
        overwrite (bool, optional): Whether the overwrite the file if
        it exists. Defaults to False.

    Returns:
        bool: The successfulness of the operation.
    """
    if not overwrite and os.path.exists(full_path):
        print("File `%s` already exists, not overwriting." % full_path)
        return True

    try:
        with open(full_path, "w") as f:
            f.write(content)
        print("Asset saved into `%s`." % full_path)
        return True

    except Exception as e:
        print("Failed to save the asset into `%s`. Error message: `%s`" % (
            full_path,
            e
            )
        )
        return False


def save_assets(misty: Callable, assets: Dict, location: str) -> List[str]:
    """Saves multiple assets.

    Args:
        misty (Callable): an instance of the misty2py.robot's Misty class.
        assets (Dict): The dictionary of assets; a list of file names under
        their asset type as the keyword.
        location (str): The absolute path to the directory where the assets
        are saved.

    Returns:
        List[str]: A list of file names for the files that failed to be saved.
    """
    failed_list = []
    action = "_file"
    for asset_type, files in assets.items():
        for file in files:
            params, name, ext = get_asset_properties(asset_type, file)
            response = misty.get_info(asset_type + action, params=params)
            result = response.get("result")

            if not result:
                failed_list.append(file)

            else:
                file_name = "%s_%s_%s_in_base64.txt" % (asset_type, name, ext)
                full_path = os.path.join(location, file_name)
                file_content = result.get("base64")
                if not file_content:
                    failed_list.append(file)
                else:
                    success = save_base64_str(full_path, file_content)
                    if not success:
                        failed_list.append(file)

    return failed_list


def delete_assets(
    misty: Callable,
    assets: Dict,
    ignore_list: List = []
) -> List[str]:
    """Deletes all non-system files except those that are in the ignore list.

    Args:
        misty (Callable): an instance of the misty2py.robot's Misty class.
        assets (Dict): The dictionary of assets; a list of file names under
        their asset type as the keyword.
        ignore_list (List, optional): The list of file names for files that
        should not be deleted. Defaults to [].

    Returns:
        List[str]: The list of files that were succesfully deleted.
    """
    action = "_delete"
    delete_list = []

    for asset_type, files in assets.items():
        for file in files:
            if file not in ignore_list:
                if asset_type == "recording":
                    data = {"Name": file}
                else:
                    data = {"FileName": file}
                response = misty.perform_action(asset_type + action, data=data)
                status = response.get("status")
                if status:
                    if status == "Success":
                        print("Successfully deleted the asset `%s`." % file)
                        delete_list.append(file)
                    else:
                        print(
                            "Failed to delete the asset `%s`. Message: `%s`"
                            % (file, response)
                        )

    return delete_list


def free_memory(
    misty: Callable,
    assets: List = ["audio", "image", "video", "recording"],
    save: bool = True,
    save_dir: str = "data",
) -> Dict:
    """Removes all non-system files in Misty's memory.

    Args:
        misty (Callable): an instance of the misty2py.robot's Misty class.
        assets (List, optional): The list of asset types to delete. Defaults
        to ["audio", "image", "video", "recording"].
        save (bool, optional): Whether to save the files that are being
        deleted from Misty's memory. Defaults to True.
        save_dir (str, optional): The path to the directory that will store
        saved assets. Defaults to "data".

    Returns:
        Dict: The dictionary with keys `overall_success` for the success
        of the entire operation and keys corresponding to the actions within
        the operation with successfulness of the action and messages about
        the action.
    """
    save_dir = get_abs_path(save_dir)
    enable_audio = misty.perform_action("audio_enable")
    enable_av = misty.perform_action("streaming_av_enable")
    enable_camera = misty.perform_action("camera_enable")

    assets_to_delete = get_non_system_assets(misty, assets=assets)
    deletion = {}

    if len(assets_to_delete) == 0:
        deletion["status"] = "Success"
        deletion["message"] = "No non-system files present."

    else:
        failed_to_save_list = []

        if save:
            failed_to_save_list = save_assets(
                misty,
                assets_to_delete,
                save_dir
            )

        deleted = delete_assets(misty, assets_to_delete, failed_to_save_list)

        if len(deleted) > 0:
            deletion["status"] = "Success"
            deletion["message"] = "Successfully deleted \
                following assets: %s" % str(deleted)

        else:
            deletion["status"] = "Failed"
            deletion["message"] = "Failed to delete any assets."

    disable_audio = misty.perform_action("audio_disable")
    disable_av = misty.perform_action("streaming_av_disable")
    disable_camera = misty.perform_action("camera_disable")

    return success_parser_from_dicts(
        enable_audio=enable_audio,
        enable_av=enable_av,
        enable_camera=enable_camera,
        deletion=deletion,
        disable_audio=disable_audio,
        disable_av=disable_av,
        disable_camera=disable_camera,
    )


if __name__ == "__main__":
    from misty2py.utils.utils import get_misty

    print(free_memory(get_misty(), "data"))



"""Colour related utility functions."""
from typing import Any, Dict, Union

from misty2py.utils.utils import query_dict_with_fallback


def parse_rgb_value(value: Union[int, str, None]) -> int:
    """Parses and validates an rgb value.

    Args:
        value (Union[int, str, None]): an rgb value.

    Raises:
        TypeError: if the value is not int or str parseable to int.
        ValueError: if the value is int or str parseable to int but is not
        a valid rgb value.

    Returns:
        int: the rgb value as an int if parseable and valid.
    """
    try:
        value = int(value)
    except Exception:
        raise TypeError("An rgb value must be int or string parseable to int.")

    if not (value >= 0 and value <= 255):
        raise ValueError("An rgb value must be \
            between 0 and 255 (bounds including)")

    return value


def rgb_values_to_dict(
    red: Union[int, str], green: Union[int, str], blue: Union[int, str]
) -> Dict:
    """Returns rgb dictionary in the form required by Misty API from rgb
    values supplied as str or int."""
    return {
        "red": parse_rgb_value(red),
        "green": parse_rgb_value(green),
        "blue": parse_rgb_value(blue),
    }


def validate_rgb_dict(potential_rgb: Dict) -> Dict:
    """Validates a potentially rgb dictionary. Returns the rgb dictionary if
    successful."""
    return {
        "red": parse_rgb_value(potential_rgb.get("red")),
        "green": parse_rgb_value(potential_rgb.get("green")),
        "blue": parse_rgb_value(potential_rgb.get("blue")),
    }


def get_rgb_from_unknown(potential_rgb: Any, allowed_data: Dict = {}) -> Dict:
    """Attempts to parse an rgb dictionary as required by Misty API
    from potential_rgb.

    Args:
        potential_rgb (Any): any data that might be an rgb dict.
        allowed_data (Dict, optional): the dictionary of allowed data
        shortcuts (string keys, dict values). Defaults to {}.

    Raises:
        TypeError: if the type of potential_rgb is not dict or parseable
        to dict.

    Returns:
        Dict: rgb dictionary as required by Misty API.
    """
    if isinstance(potential_rgb, str):
        potential_rgb = allowed_data.get(potential_rgb)

    if isinstance(potential_rgb, Dict):
        return validate_rgb_dict(potential_rgb)

    else:
        raise TypeError(
            "Incorrect type: `%s`, expecting Dict." % type(potential_rgb))


def construct_transition_dict(
    data: Dict,
    allowed_data: Dict,
    fallback_time: int = 500,
    fallback_trans: str = "Breathe",
) -> Dict:
    """Constructs a transition dictionary in the form required by Misty API.

    Args:
        data (Dict): the dictionary to transform, must have keys col1 and col2
        that contain a string (data shortcut for a colour) or a dict (rgb
        dictionary) and should have keys transition and time.
        allowed_data (Dict): the dictionary of allowed data shortcuts (string
        keys, dict values).
        fallback_time (int, optional): The default time between colours
        switching in ms. Defaults to 500.
        fallback_trans (str, optional): The default transition type, can be
        "Blink", "Breathe" or "TransitOnce". Defaults to "Breathe".

    Returns:
        Dict: The input dictionary in the form required by Misty API's
        endpoint /api/led/transition.
    """
    col1 = get_rgb_from_unknown(data.get("col1"), allowed_data=allowed_data)
    col2 = get_rgb_from_unknown(data.get("col2"), allowed_data=allowed_data)

    return {
        "Red": col1.get("red"),
        "Green": col1.get("green"),
        "Blue": col1.get("blue"),
        "Red2": col2.get("red"),
        "Green2": col2.get("green"),
        "Blue2": col2.get("blue"),
        "TransitionType": query_dict_with_fallback(
            data, "transition", fallback_trans, required_type=str
        ),
        "TimeMS": query_dict_with_fallback(
            data, "time", fallback_time, required_type=int
        ),
    }



"""This script implements the action keywords matching to Misty's API
endpoints, sending action requests and data shortcuts.
"""
import json
import sys
from os import path
from typing import Dict, Union

import requests

from misty2py.utils.colours import construct_transition_dict

this_directory = path.abspath(path.dirname(__file__))
ACTIONS_JSON = str(path.join(this_directory, "allowed_actions.json"))
DATA_JSON = str(path.join(this_directory, "allowed_data.json"))


class Post:
    """A class representing the POST url request method.

    Attributes:
        ip (str): The IP address where the requests are sent.
        allowed_actions (dict): The dictionary of custom action keywords
        matching to the Misty's API endpoints.
        allowed_data (dict): The dictionary of custom data shortcuts matching
        to the json dictionaries required my Misty's API.
    """

    def __init__(
        self,
        ip: str,
        custom_allowed_actions: Dict = {},
        custom_allowed_data: Dict = {}
    ) -> None:
        """Initialises a Post object.

        Args:
            ip (str): The IP address where the requests are sent.
            custom_allowed_actions (Dict, optional): The dictionary of action
            keywords. Defaults to {}.
            custom_allowed_data (Dict, optional): The dictionary of data
            shortcuts. Defaults to {}.
        """

        self.ip = ip

        allowed_actions = custom_allowed_actions
        f = open(ACTIONS_JSON)
        allowed_actions.update(json.loads(f.read()))
        f.close()
        self.allowed_actions = allowed_actions

        allowed_data = custom_allowed_data
        f = open(DATA_JSON)
        allowed_data.update(json.loads(f.read()))
        f.close()
        self.allowed_data = allowed_data

    def perform_action(
        self, endpoint: str, data: Dict, request_method: str = "post"
    ) -> Dict:
        """Sends a POST request.

        Args:
            endpoint (str): The API endpoint to which the request is sent.
            data (Dict): The json data supplied in the body of the request.
            request_method (str, optional): The request method.
            Defaults to "post".

        Returns:
            Dict: The API response.
        """

        if request_method == "post":
            response = requests.post(
                "http://%s/%s" % (self.ip, endpoint),
                json=data
            )
        else:
            response = requests.delete(
                "http://%s/%s" % (self.ip, endpoint),
                json=data
            )
        try:
            return response.json()
        except Exception as e:
            return {
                "status": "Success",
                "content": response.content,
                "error_msg": e
            }


class Action(Post):
    """A class representing an action request for Misty.
    A subclass of Post()."""

    def perform_action(
        self, action_name: str, data: Union[str, Dict], data_method: str
    ) -> Dict:
        """Sends an action request to Misty.

        Args:
            action_name (str): The action keyword specifying which action
            is requested.
            data (Union[str, Dict]): The data shortcut representing the data
            supplied in the body of the request or the json dictionary
            to be supplied in the body of the request.
            data_method (str): "dict" if the data is supplied as a json
            dictionary, "string" if the data is supplied as a data shortcut.

        Returns:
            Dict: the response from the API.
        """

        if action_name not in self.allowed_actions.keys():
            return {
                "status": "Failed",
                "message": "Command `%s` not supported." % action_name,
            }

        else:
            if data_method == "dict":
                try:
                    return super().perform_action(
                        self.allowed_actions[action_name]["endpoint"],
                        data,
                        request_method=(
                            self.allowed_actions[action_name]["method"]
                        )
                    )

                except Exception as e:
                    return {
                        "status": "Failed",
                        "message": "Error occured. Error message: `%s`." % e,
                    }

            elif data_method == "string" and data in self.allowed_data:
                try:
                    return super().perform_action(
                        self.allowed_actions[action_name]["endpoint"],
                        self.allowed_data[data],
                        request_method=(
                            self.allowed_actions[action_name]
                                                ["method"]
                        ),
                    )

                except Exception as e:
                    return {
                        "status": "Failed",
                        "message": "Error occured. Error message: `%s`" % e,
                    }

            else:
                return {
                    "status": "Failed",
                    "message": "Data shortcut `%s` not supported." % data,
                }

    def action_handler(self, action_name: str, data: Union[Dict, str]) -> Dict:
        """Sends Misty a request to perform an action.

        Args:
            action_name (str): The keyword specifying the action to perform.
            data (Union[Dict, str]): The data to send in the request body
            in the form of a data shortcut or a json dictionary.

        Returns:
            Dict: response from the API
        """

        if (
            action_name == "led_trans" and
            isinstance(data, Dict) and
            len(data) >= 2 and
            len(data) <= 4
        ):

            try:
                data = construct_transition_dict(data, self.allowed_data)

            except ValueError as e:
                return {
                    "status": "Failed",
                    "message": "The data is not in correct format.",
                    "details": e,
                }

        data_method = ""

        if isinstance(data, Dict):
            data_method = "dict"

        else:
            data_method = "string"

        return self.perform_action(action_name, data, data_method)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['misty2py', 'misty2py.basic_skills', 'misty2py.utils']

package_data = \
{'': ['*']}

install_requires = \
['python-dotenv>=0.17.1,<0.18.0',
 'requests>=2.25.1,<3.0.0',
 'websocket-client>=0.58.0,<0.59.0']

setup_kwargs = {
    'name': 'misty2py',
    'version': '4.1.5',
    'description': 'A Python library for Misty II development',
    'long_description': '# Misty2py\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black) [![GitHub license](https://img.shields.io/github/license/Naereen/StrapDown.js.svg)](https://github.com/ChrisScarred/misty2py/blob/main/LICENSE)\n\nMisty2py is a Python 3 library for Misty II development using [Misty API](https://docs.mistyrobotics.com/misty-ii/rest-api/api-reference/ "Misty Robotics REST API").\n\n## Features\n\nMisty2py can be used to develop complex skills (behaviours) for the Misty II robot utilising:\n\n- **actions** via sending a `POST` or `DELETE` requests to Misty\'s API;\n- **informations** via sending a `GET` request to Misty\'s API;\n- **continuous streams of data** via subscribing to event types on Misty\'s websockets.\n\nMisty2py uses following concepts for easy of usage:\n\n- **action keywords** - customisable python-styled keywords for endpoints of Misty\'s API that correspond to performing actions;\n- **information keywords** - customisable python-styled keywords for endpoints of Misty\'s API that correspond to retrieving information;\n- **data shortcuts** - customisable python-styled keywords for commonly used data that are supplied to Misty\'s API as the body of a `POST` request.\n\n## Usage\n\n### Getting started\n\n- Start by making **a new instance** of `misty2py.robot`\'s `Misty` by `misty_robot = Misty("ip_address_here")`\n  - Substitute `ip_address_here` with the IP address of your Misty.\n  - `misty2py.utils.env_loader` module contains `EnvLoader` class that can be used to load Misty\'s IP from the .env file in your project\'s home directory:\n    - Create the `.env` file and write `MISTY_IP_ADDRESS="[ip_address_here]"` in it.\n    - Initialise an `EnvLoader` object via `env_loader = EnvLoader()`; this loads the environment variables.\n    - Use `env_loader.get_ip()` to obtain the IP address.\n- Use the method `misty_robot.perform_action()` to tell Misty to **perform an action**.\n- Use the method `misty_robot.get_info()` to tell Misty to **return information**.\n- Use the method `misty_robot.event()` to initialise, obtain and stop **continuous streams of data** from Misty\'s event types.\n\n### Obtaining information\n\nObtaining digital information is handled by `misty2py.robot::get_info` method.\n\n`misty2py.robot::get_info` has following arguments:\n\n- `info_name` - *required;* the string information keyword corresponding to an endpoint in Misty\'s API;\n- `params` - *optional;* a dictionary of parameter name and parameter value pairs, defaults to `{}`.\n\n### Performing actions\n\nPerforming physical and digital actions including removal of non-system files is handled by `misty2py.robot::perform_action()` method.\n\n`misty2py.robot::perform_action()` has following arguments:\n\n- `action_name` - *required;* the string action keyword corresponding to an endpoint in Misty\'s API;\n- `data` - *optional;* the data to pass to the request as a dictionary or a data shortcut (string), defaults to `{}`.\n\n### Event types\n\nTo obtain event data in Misty\'s framework, it is required to **subscribe** to an event type on Misty\'s websocket server. Misty\'s websocket server then streams data to the websocket client, in this implementation via a separate thread. To **access this data,** `misty2py.robot::event` method must be called with `"get_data"` parameter from the main thread. When data are no longer required to be streamed to the client, an event type can be **unsubscribed** to kill the event thread.\n\n#### Subscription\n\nSubscribe to an event via `misty2py.robot::event` with the parameter `"subscribe"` and following keyword arguments:\n\n    - `type` - *required;* event type string as documented in [Event Types Docs](https://docs.mistyrobotics.com/misty-ii/robot/sensor-data/ "Misty Robotics Event Types").\n    - `name` - *optional;* a custom event name string; must be unique.\n    - `return_property` - *optional;* the property to return from Misty\'s websockets; all properties are returned if return_property is not supplied.\n    - `debounce` - *optional;* the interval in ms at which new information is sent; defaults to 250.\n    - `len_data_entries` - *optional;* the maximum number of data entries to keep (discards in fifo style); defaults to 10.\n    - `event_emitter` - *optional;* an event emitter function which emits an event upon message recieval. Supplies the message content as an argument.\n\n#### Accessing the data and the log\n\nAccess the data of an event or its log via `misty2py.robot::event` with the parameter `"get_data"` or `"get_log"` and a keyword argument `name` (the name of the event).\n\n#### Unsubscribing\n\nUnsubscribe from an event via `misty2py.robot::event` with the parameter `"unsubscribe"` and a keyword argument `name` (the name of the event).\n\n#### Basic example\n\n```python\nimport time\n\nfrom misty2py.robot import Misty\nfrom misty2py.utils.env_loader import EnvLoader\n\nenv_loader = EnvLoader\n\nm = Misty(env_loader.get_ip())\n\nd = m.event("subscribe", type = "BatteryCharge")\ne_name = d.get("event_name")\n\ntime.sleep(1)\n\nd = m.event("get_data", name = e_name)\n\nd = m.event("unsubscribe", name = e_name)\n```\n\n#### Event emitter usage - example\n\n```python\nimport time\nfrom pymitter import EventEmitter\n\nfrom misty2py.robot import Misty\nfrom misty2py.utils.env_loader import EnvLoader\n\nenv_loader = EnvLoader\n\nm = Misty(env_loader.get_ip())\nee = EventEmitter()\nevent_name = "myevent_001"\n\n@ee.on(event_name)\ndef listener(data):\n    print(data)\n\nd = m.event("subscribe", type = "BatteryCharge", name = event_name, event_emitter = ee)\n\ntime.sleep(2)\n\nd = m.event("unsubscribe", name = event_name)\n```\n\n### Utilities\n\nThe sub-package `misty2py.utils` contains utility modules, including:\n\n- `misty2py.utils.base64` - contains base64 encoders and decoders.\n  - `misty2py.utils.base64::content_to_base64` - encodes a string or a file into base64 string and either returns the encoded string or saves the encoded string into a file.\n  - `misty2py.utils.base64::base64_to_content` - decodes a string or a file from base64 and either returns the decoded string or saves the decoded file.\n\n- `misty2py.utils.colours` - contains colour-related utility functions.\n  - `misty2py.utils.colours::parse_rgb_value` - parses and validates a potentially rgb value (a string or an integer).\n  - `misty2py.utils.colours::rgb_values_to_dict` - turns integer or string triple of rgb values into the rgb dictionary form required by Misty API.\n  - `misty2py.utils.colours::validate_rgb_dict` - validates an rgb dictionary.\n  - `misty2py.utils.colours::get_rgb_from_unknown` - attempts to obtain an rgb dictionary from an unknown data content (possibly a data shortcut or a dictionary)\n  - `misty2py.utils.colours::construct_transition_dict` - constructs the transition dictionary as required by Misty API from a dictionary of two colours supplied either as data shortcuts or rgb dictionaries.\n\n- `misty2py.utils.env_loader` contains the `EnvLoader` class which can be used to load environmental variables, specifically the IP address of the Misty robot.\n  - `misty2py.utils.env_loader::EnvLoader.__init__` (aka `misty2py.utils.env_loader.EnvLoader()`) - initialises the `EnvLoader`, takes optional argument `env_path` (str) for a custom path to the `.env` file.\n  - `misty2py.utils.env_loader::EnvLoader.get_ip` - obtains the value of the environmental variable `MISTY_IP_ADDRESS` if set, else returns `None`.\n\n- `misty2py.utils.generators` - contains generator functions.\n  - `misty2py.utils.generators::get_random_string` - returns a random `n` characters long string of ASCII letters and digits\n\n- `misty2py.utils.messages` - contains message parsing methods.\n  - `misty2py.utils.messages::message_parser` - parses a message from misty2py json Dict reply to a string.\n  - `misty2py.utils.messages::compose_str` - adds `potential_str` to `main_str` if `potential_str` is not `None`, else adds `fallback` to `main_str` if `fallback` is not `None`, else returns `main_str`\n\n- `misty2py.utils.utils` - contains generally useful functions that do not fall into any of the above categories.\n  - `misty2py.utils.utils::query_dict_with_fallback` - safely queries a dictionary and if the key does not exist or the value has inappropriate type, returns the fallback value.\n\n### Keywords and shortcuts\n\n<details> \n  <summary>List of supported action keywords</summary>\n  \n- `led` for **post** request to `api/led` endpoint\n- `led_trans` for **post** request to `api/led/transition` endpoint\n- `notification_settings` for **post** request to `api/notification/settings` endpoint\n- `audio_upload` for **post** request `api/audio` to endpoint\n- `audio_play` for **post** request to `api/audio/play` endpoint\n- `audio_pause` for **post** request to `api/audio/pause` endpoint\n- `audio_stop` for **post** request to `api/audio/stop` endpoint\n- `audio_delete` for **delete** request to `api/audio` endpoint\n- `audio_record_start` for **post** request to `api/audio/record/start` endpoint\n- `audio_record_stop` for **post** request to `api/audio/record/stop` endpoint\n- `audio_disable` for **post** request to `api/services/audio/disable` endpoint\n- `audio_enable` for **post** request to `api/services/audio/enable` endpoint\n- `image_upload` for **post** request to `api/images` endpoint\n- `image_show` for **post** request to `api/images/display` endpoint\n- `image_settings` for **post** request to `api/images/settings` endpoint\n- `image_delete` for **delete** request to `api/images` endpoint\n- `text_show` for **post** request to `api/text/display` endpoint\n- `text_settings` for **post** request to `api/text/settings` endpoint\n- `video_upload` for **post** request to `api/videos` endpoint\n- `video_show` for **post** request to `api/videos/display` endpoint\n- `video_settings` for **post** request to `api/videos/settings` endpoint\n- `video_delete` for **delete** request to `api/videos` endpoint\n- `blink_mapping_delete` for **delete** request to `api/blink/images` endpoint\n- `slam_enable` for **post** request to `api/services/slam/enable` endpoint\n- `slam_disable` for **post** request to `api/services/slam/disable` endpoint\n- `slam_sensors_reset` for **post** request to `api/slam/reset` endpoint\n- `slam_mapping_start` for **post** request to `api/slam/map/start` endpoint\n- `slam_mapping_stop` for **post** request to `api/slam/map/stop` endpoint\n- `slam_map_current` for **post** request to `api/slam/map/current` endpoint\n- `slam_map_rename` for **post** request to `api/slam/map/rename` endpoint\n- `slam_infrared_settings` for **post** request to `api/slam/settings/ir` endpoint\n- `slam_visible_settings` for **post** request to `api/slam/settings/visible` endpoint\n- `slam_map_delete` for **delete** request to `api/slam/map` endpoint\n- `slam_docking_locate_start` for **post** request to `api/slam/docking/start` endpoint\n- `slam_docking_locate_stop` for **post** request to `api/slam/docking/stop` endpoint\n- `streaming_slam_start` for **post** request to `api/slam/streaming/start` endpoint\n- `streaming_slam_stop` for **post** request to `api/slam/streaming/stop` endpoint\n- `slam_track_start` for **post** request to `api/slam/track/start` endpoint\n- `slam_track_stop` for **post** request to `api/slam/track/stop` endpoint\n- `recording_start` for **post** request to `api/videos/recordings/start` endpoint\n- `recording_stop` for **post** request to `api/videos/recordings/stop` endpoint\n- `recording_rename` for **post** request to `api/videos/recordings/rename` endpoint\n- `recording_delete` for **delete** request to `api/videos/recordings` endpoint\n- `face_detection_start` for **post** request to `api/faces/detection/start` endpoint\n- `face_detection_stop` for **post** request to `api/faces/detection/stop` endpoint\n- `face_recognition_start` for **post** request to `api/faces/recognition/start` endpoint\n- `face_recognition_stop` for **post** request to `api/faces/recognition/stop` endpoint\n- `face_train_start` for **post** request to `api/faces/training/start` endpoint\n- `face_train_cancel` for **post** request to `api/faces/training/cancel` endpoint\n- `face_delete` for **delete** request to `api/faces` endpoint\n- `skill_upload` for **post** request to `api/skills` endpoint\n- `skill_start` for **post** request to `api/skills/start` endpoint\n- `skills_reload` for **post** request to `api/skills/reload` endpoint\n- `skill_load` for **post** request to `api/skills/load` endpoint\n- `skill_cancel` for **post** request to `api/skills/cancel` endpoint\n- `skill_delete` for **delete** request to `api/skills` endpoint\n- `wifi_add` for **post** request to `api/networks/create` endpoint\n- `wifi_connect` for **post** request to `api/networks` endpoint\n- `wifi_delete` for **delete** request to `api/networks` endpoint\n- `wifi_hotspot_start` for **post** request to `api/networks/hotspot/start` endpoint\n- `wifi_hotspot_stop` for **post** request to `api/networks/hotspot/stop` endpoint\n- `write_serial` for **post** request to `api/serial` endpoint\n- `event_listener` for **post** request to `api/skills/event` endpoint\n- `website_show` for **post** request to `api/webviews/display` endpoint\n- `website_settings` for **post** request to `api/webviews/settings` endpoint\n- `blink_on` for **post** request to `api/blink` endpoint\n- `blink_settings` for **post** request to `api/blink/settings` endpoint\n- `display_settings` for **post** request to `api/display/settings` endpoint\n- `flashlight_on` for **post** request to `api/flashlight` endpoint\n- `speak` for **post** request to `api/tts/speak` endpoint\n- `speak_stop` for **post** request to `api/tts/stop` endpoint\n- `speech_capture` for **post** request to `api/audio/speech/capture` endpoint\n- `drive` for **post** request to `api/drive` endpoint\n- `drive_arc` for **post** request to `api/drive/arc` endpoint\n- `drive_heading` for **post** request to `api/drive/hdt` endpoint\n- `drive_time` for **post** request to `api/drive/time` endpoint\n- `drive_track` for **post** request to `api/drive/track` endpoint\n- `drive_stop` for **post** request to `api/drive/stop` endpoint\n- `drive_to_loc` for **post** request to `api/drive/coordinates` endpoint\n- `drive_on_path` for **post** request to `api/drive/path` endpoint\n- `halt` for **post** request to `api/halt` endpoint\n- `arm_move` for **post** request to `api/arms` endpoint\n- `arms_move` for **post** request to `api/arms/set` endpoint\n- `head_move` for **post** request to `api/head` endpoint\n- `hazard_settings` for **post** request to `api/hazard/updatebasesettings` endpoint\n- `streaming_av_start` for **post** request to `api/avstreaming/start` endpoint\n- `streaming_av_stop` for **post** request to `api/avstreaming/stop` endpoint\n- `streaming_av_disable` for **post** request to `api/services/avstreaming/disable` endpoint\n- `streaming_av_enable` for **post** request to `api/services/avstreaming/enable` endpoint\n- `keyphrase_recognition_start` for **post** request to `api/audio/keyphrase/start` endpoint\n- `keyphrase_recognition_stop` for **post** request to `api/audio/keyphrase/stop` endpoint\n- `update_allow` for **post** request to `api/system/update/allow` endpoint\n- `update_perform` for **post** request to `api/system/update` endpoint\n- `update_perform_targeted` for **post** request to `api/system/update/component` endpoint\n- `update_prevent` for **post** request to `api/system/update/prevent` endpoint\n- `error_text_clear` for **post** request to `api/text/error/clear` endpoint\n- `camera_disable` for **post** request to `api/services/camera/disable` endpoint\n- `camera_enable` for **post** request to `api/services/camera/enable` endpoint\n- `restart` for **post** request to `api/reboot` endpoint\n- `volume_settings` for **post** request to `api/audio/volume` endpoint\n- `logs_settings` for **post** request to `api/logs/level` endpoint\n- `websocket_settings` for **post** request to `api/websocket/version` endpoint\n- `external_request` for **post** request to `api/request` endpoint\n</details>\n\n<details> \n    <summary>List of supported information keywords</summary>\n\n- `audio_file` for **get** request to `api/audio` endpoint\n- `audio_list` for **get** request to `api/audio/list` endpoint\n- `audio_status` for **get** request to `api/services/audio` endpoint\n- `image_file` for **get** request to `api/images` endpoint\n- `image_list` for **get** request to `api/images/list` endpoint\n- `video_file` for **get** request to `api/videos` endpoint\n- `video_list` for **get** request to `api/videos/list` endpoint\n- `av_status` for **get** request to `api/services/avstreaming` endpoint\n- `sensor_values` for **get** request to `api/serial` endpoint\n- `map_file` for **get** request to `api/slam/map` endpoint\n- `current_map_id` for **get** request to `api/slam/map/current` endpoint\n- `map_id_list` for **get** request to `api/slam/map/ids` endpoint\n- `slam_diagnostics` for **get** request to `api/slam/diagnostics` endpoint\n- `slam_path` for **get** request to `api/slam/path` endpoint\n- `slam_status` for **get** request to `api/slam/status` endpoint\n- `slam_enabled` for **get** request to `api/services/slam` endpoint\n- `picture_depth` for **get** request to `api/cameras/depth` endpoint\n- `picture_fisheye` for **get** request to `api/cameras/fisheye` endpoint\n- `picture_rgb` for **get** request to `api/cameras/rgb` endpoint\n- `faces_known` for **get** request to `api/faces` endpoint\n- `recording_file` for **get** request to `api/videos/recordings` endpoint\n- `recording_list` for **get** request to `api/videos/recordings/list` endpoint\n- `skills_running` for **get** request to `api/skills/running` endpoint\n- `skills_known` for **get** request to `api/skills` endpoint\n- `wifis_available` for **get** request to `api/networks/scan` endpoint\n- `wifis_saved` for **get** request to `api/networks` endpoint\n- `battery_status` for **get** request to `api/battery` endpoint\n- `camera_status` for **get** request to `api/services/camera` endpoint\n- `blink_settings` for **get** request to `api/blink/settings` endpoint\n- `hazards_settings` for **get** request to `api/hazards/settings` endpoint\n- `camera_settings` for **get** request to `api/camera` endpoint\n- `slam_visible_settings` for **get** request to `api/slam/settings/visible` endpoint\n- `slam_infrared_settings` for **get** request to `api/slam/settings/ir` endpoint\n- `update_settings` for **get** request to `api/system/update/settings` endpoint\n- `device` for **get** request to `api/device` endpoint\n- `help` for **get** request to `api/help` endpoint\n- `log` for **get** request to `api/logs` endpoint\n- `log_level` for **get** request to `api/logs/level` endpoint\n- `update_available` for **get** request to `api/system/updates` endpoint\n- `websockets` for **get** request to `api/websockets` endpoint\n- `websocket_version` for **get** request to `api/websocket/version`\n</details>\n\n<details> \n    <summary>List of supported data shortcuts</summary>\n\n- `led_off` for `{ "red": "0", "green": "0", "blue": "0" }`\n- `white_light` for `{ "red": "255", "green": "255", "blue": "255" }`\n- `red_light` for `{ "red": "255", "green": "0", "blue": "0" }`\n- `green_light` for `{ "red": "0", "green": "255", "blue": "0" }`\n- `blue_light` for `{ "red": "0", "green": "0", "blue": "255" }`\n- `yellow_light` for `{ "red": "255", "green": "255", "blue": "0" }`\n- `cyan_light` for `{ "red": "0", "green": "255", "blue": "255" }`\n- `magenta_light` for `{ "red": "255", "green": "0", "blue": "255" }`\n- `orange_light` for `{ "red": "255", "green": "125", "blue": "0" }`\n- `lime_light` for `{ "red": "125", "green": "255", "blue": "0" }`\n- `aqua_light` for `{ "red": "0", "green": "255", "blue": "125" }`\n- `azure_light` for `{ "red": "0", "green": "125", "blue": "255" }`\n- `violet_light` for `{ "red": "125", "green": "0", "blue": "255" }`\n- `pink_light` for `{ "red": "255", "green": "0", "blue": "125" }`\n- `low_volume` for `{ "Volume": "5" }`\n- `image_admiration` for `{"FileName": "e_Admiration.jpg"}`\n- `image_aggressiveness` for `{"FileName": "e_Aggressiveness.jpg"}`\n- `image_amazement` for `{"FileName": "e_Amazement.jpg"}`\n- `image_anger` for `{"FileName": "e_Anger.jpg"}`\n- `image_concerned` for `{"FileName": "e_ApprehensionConcerned.jpg"}`\n- `image_contempt` for `{"FileName": "e_Contempt.jpg"}`\n- `image_content_left` for `{"FileName": "e_ContentLeft.jpg"}`\n- `image_content_right` for `{"FileName": "e_ContentRight.jpg"}`\n- `image_content_default` for `{"FileName": "e_DefaultContent.jpg"}`\n- `image_disgust` for `{"FileName": "e_Disgust.jpg"}`\n- `image_disoriented` for `{"FileName": "e_Disoriented.jpg"}`\n- `image_hilarious` for `{"FileName": "e_EcstacyHilarious.jpg"}`\n- `image_starry_eyed` for `{"FileName": "e_EcstacyStarryEyed.jpg"}`\n- `image_fear` for `{"FileName": "e_Fear.jpg"}`\n- `image_grief` for `{"FileName": "e_Grief.jpg"}`\n- `image_joy_1` for `{"FileName": "e_Joy.jpg"}`\n- `image_joy_2` for `{"FileName": "e_Joy2.jpg"}`\n- `image_goofy_1` for `{"FileName": "e_JoyGoofy.jpg"}`\n- `image_goofy_2` for `{"FileName": "e_JoyGoofy2.jpg"}`\n- `image_goofy_3` for `{"FileName": "e_JoyGoofy3.jpg"}`\n- `image_love` for `{"FileName": "e_Love.jpg"}`\n- `image_rage_1` for `{"FileName": "e_Rage.jpg"}`\n- `image_rage_2` for `{"FileName": "e_Rage2.jpg"}`\n- `image_rage_3` for `{"FileName": "e_Rage3.jpg"}`\n- `image_rage_4` for `{"FileName": "e_Rage4.jpg"}`\n- `image_remorse` for `{"FileName": "e_RemorseShame.jpg"}`\n- `image_sadness` for `{"FileName": "e_Sadness.jpg"}`\n- `image_sleping_1` for `{"FileName": "e_Sleeping.jpg"}`\n- `image_sleeping_2` for `{"FileName": "e_SleepingZZZ.jpg"}`\n- `image_sleepy_1` for `{"FileName": "e_Sleepy.jpg"}`\n- `image_sleepy_2` for `{"FileName": "e_Sleepy2.jpg"}`\n- `image_sleepy_3` for `{"FileName": "e_Sleepy3.jpg"}`\n- `image_sleepy_4` for `{"FileName": "e_Sleepy4.jpg"}`\n- `image_surprise` for `{"FileName": "e_Surprise.jpg"}`\n- `image_system_black_screen` for `{"FileName": "e_SystemBlackScreen.jpg"}`\n- `image_system_blink_large` for `{"FileName": "e_SystemBlinkLarge.jpg"}`\n- `image_system_blink_standard` for `{"FileName": "e_SystemBlinkStandard.jpg"}`\n- `image_system_camera` for `{"FileName": "e_SystemCamera.jpg"}`\n- `image_system_flash` for `{"FileName": "e_SystemFlash.jpg"}`\n- `image_system_gear_prompt` for `{"FileName": "e_SystemGearPrompt.jpg"}`\n- `image_system_logo_prompt` for `{"FileName": "e_SystemLogoPrompt.jpg"}`\n- `image_terror_1` for `{"FileName": "e_Terror.jpg"}`\n- `image_terror_2` for `{"FileName": "e_Terror2.jpg"}`\n- `image_terror_left` for `{"FileName": "e_TerrorLeft.jpg"}`\n- `image_terror_right` for `{"FileName": "e_TerrorRight.jpg"}`\n- `sound_acceptance` for `{ "FileName": "s_Acceptance.wav" }`\n- `sound_amazement_1` for `{ "FileName": "s_Amazement.wav" }`\n- `sound_amazement_2` for `{ "FileName": "s_Amazement2.wav" }`\n- `sound_anger_1` for `{ "FileName": "s_Anger.wav" }`\n- `sound_anger_2` for `{ "FileName": "s_Anger2.wav" }`\n- `sound_anger_3` for `{ "FileName": "s_Anger3.wav" }`\n- `sound_anger_4` for `{ "FileName": "s_Anger4.wav" }`\n- `sound_annoyance_1` for `{ "FileName": "s_Annoyance.wav" }`\n- `sound_annoyance_2` for `{ "FileName": "s_Annoyance2.wav" }`\n- `sound_annoyance_3` for `{ "FileName": "s_Annoyance3.wav" }`\n- `sound_annoyance_4` for `{ "FileName": "s_Annoyance4.wav" }`\n- `sound_awe_1` for `{ "FileName": "s_Awe.wav" }`\n- `sound_awe_2` for `{ "FileName": "s_Awe2.wav" }`\n- `sound_awe_3` for `{ "FileName": "s_Awe3.wav" }`\n- `sound_boredom` for `{ "FileName": "s_Boredom.wav" }`\n- `sound_disapproval` for `{ "FileName": "s_Disapproval.wav" }`\n- `sound_disgust_1` for `{ "FileName": "s_Disgust.wav" }`\n- `sound_disgust_2` for `{ "FileName": "s_Disgust2.wav" }`\n- `sound_disgust_3` for `{ "FileName": "s_Disgust3.wav" }`\n- `sound_disoriented_1` for `{ "FileName": "s_DisorientedConfused.wav" }`\n- `sound_disoriented_2` for `{ "FileName": "s_DisorientedConfused2.wav" }`\n- `sound_disoriented_3` for `{ "FileName": "s_DisorientedConfused3.wav" }`\n- `sound_disoriented_4` for `{ "FileName": "s_DisorientedConfused4.wav" }`\n- `sound_disoriented_5` for `{ "FileName": "s_DisorientedConfused5.wav" }`\n- `sound_disoriented_6` for `{ "FileName": "s_DisorientedConfused6.wav" }`\n- `sound_distraction` for `{ "FileName": "s_Distraction.wav" }`\n- `sound_ecstacy_1` for `{ "FileName": "s_Ecstacy.wav" }`\n- `sound_ecstacy_2` for `{ "FileName": "s_Ecstacy2.wav" }`\n- `sound_fear` for `{ "FileName": "s_Fear.wav" }`\n- `sound_grief_1` for `{ "FileName": "s_Grief.wav" }`\n- `sound_grief_2` for `{ "FileName": "s_Grief2.wav" }`\n- `sound_grief_3` for `{ "FileName": "s_Grief3.wav" }`\n- `sound_grief_4` for `{ "FileName": "s_Grief4.wav" }`\n- `sound_joy_1` for `{ "FileName": "s_Joy.wav" }`\n- `sound_joy_2` for `{ "FileName": "s_Joy2.wav" }`\n- `sound_joy_3` for `{ "FileName": "s_Joy3.wav" }`\n- `sound_joy_4` for `{ "FileName": "s_Joy4.wav" }`\n- `sound_loathing` for `{ "FileName": "s_Loathing.wav" }`\n- `sound_love` for `{ "FileName": "s_Love.wav" }`\n- `sound_phrase_bye_bye` for `{ "FileName": "s_PhraseByeBye.wav" }`\n- `sound_phrase_evil` for `{ "FileName": "s_PhraseEvilAhHa.wav" }`\n- `sound_phrase_hello` for `{ "FileName": "s_PhraseHello.wav" }`\n- `sound_phrase_no` for `{ "FileName": "s_PhraseNoNoNo.wav" }`\n- `sound_phrase_oopsy` for `{ "FileName": "s_PhraseOopsy.wav" }`\n- `sound_phrase_ow` for `{ "FileName": "s_PhraseOwOwOw.wav" }`\n- `sound_phrase_oww` for `{ "FileName": "s_PhraseOwwww.wav" }`\n- `sound_phrase_uh` for `{ "FileName": "s_PhraseUhOh.wav" }`\n- `sound_rage` for `{ "FileName": "s_Rage.wav" }`\n- `sound_sadness_1` for `{ "FileName": "s_Sadness.wav" }`\n- `sound_sadness_2` for `{ "FileName": "s_Sadness2.wav" }`\n- `sound_sadness_3` for `{ "FileName": "s_Sadness3.wav" }`\n- `sound_sadness_4` for `{ "FileName": "s_Sadness4.wav" }`\n- `sound_sadness_5` for `{ "FileName": "s_Sadness5.wav" }`\n- `sound_sadness_6` for `{ "FileName": "s_Sadness6.wav" }`\n- `sound_sadness_7` for `{ "FileName": "s_Sadness7.wav" }`\n- `sound_sleepy_1` for `{ "FileName": "s_Sleepy.wav" }`\n- `sound_sleepy_2` for `{ "FileName": "s_Sleepy2.wav" }`\n- `sound_sleepy_3` for `{ "FileName": "s_Sleepy3.wav" }`\n- `sound_sleepy_4` for `{ "FileName": "s_Sleepy4.wav" }`\n- `sound_sleepy_snore` for `{ "FileName": "s_SleepySnore.wav" }`\n- `sound_camera_shutter` for `{ "FileName": "s_SystemCameraShutter.wav" }`\n- `sound_failure` for `{ "FileName": "s_SystemFailure.wav" }`\n- `sound_success` for `{ "FileName": "s_SystemSuccess.wav" }`\n- `sound_wake` for `{ "FileName": "s_SystemWakeWord.wav" }`\n</details>\n\n#### Adding custom keywords and shortcuts\n\nCustom keywords and shortcuts can be passed to a Misty object while declaring a new instance by using the optional arguments:\n\n- `custom_info` for custom information keywords (a dictionary with keys being the information keywords and values being the endpoints),\n- `custom_actions` for custom action keywords (a dictionary with keys being the action keywords and values being a dictionary `{"endpoint" : "edpoint_value", "method" : "method_value"}` where `method_value` is either `post` or `delete`),\n- `custom_data` for custom data shortcuts (a dictionary with keys being the data shortcuts and values being the dictionary of data values).\n\nAn example:\n\n```python\ncustom_allowed_infos = {\n    "hazards_settings": "api/hazards/settings"\n}\n\ncustom_allowed_data = {\n    "amazement": {\n        "FileName": "s_Amazement.wav"\n    },\n    "red": {\n        "red": "255",\n        "green": "0",\n        "blue": "0"\n    }\n}\n\ncustom_allowed_actions = {\n    "audio_play" : {\n        "endpoint" : "api/audio/play",\n        "method" : "post"\n    },\n    "delete_audio" : {\n        "endpoint" : "api/audio",\n        "method" : "delete"\n    }\n}\n\nmisty_robot = Misty("0.0.0.0", \n    custom_info=custom_allowed_infos, \n    custom_actions=custom_allowed_actions, \n    custom_data=custom_allowed_data)\n```\n\n## Running Misty2py from source\n\n- If this is your first time using `misty2py` from source, do following:\n\n  - Copy `.env.example` to `.env`.\n  - Replace the placeholder values in the new `.env` file.\n  - Run `poetry install` to obtain all dependencies.\n\n- Run the desired script via `poetry run python -m [name]` where `[name]` is the placeholder for the module location (in Python notation).\n- If the scripts run but your Misty does not seem to respond, you have most likely provided an incorrect IP address for `MISTY_IP_ADDRESS` in `.env`.\n\n## Pytests\n\nThis repository contains several parametrised unit tests under the pytest library that can be found in the directory `tests`. The test coverage is currently 84%.\n\n### Instructions for running the tests and test coverage reports\n\n- if this is your first time using misty2py from source, follow instructions in [Running Misty2py from source](#Running-Misty2py-from-source).\n  \n- Pytests can be run via `poetry run pytest .`.\n- The coverage report can be obtained via `poetry run pytest --cov-report html --cov=misty2py tests` for HTML output or via `poetry run pytest --cov=misty2py tests` for terminal output.\n',
    'author': "Chris 'Scar(R)ed' Vajdík",
    'author_email': 'scarred.chris@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ChrisScarred/misty2py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Message parsing utility functions."""
from typing import Dict, List, Optional, Tuple


def message_parser(
    data: Dict,
    success_message: str = "Operation successful.",
    fail_message: str = "Operation failed.",
) -> str:
    """Parses a message from misty2py json Dict reply to a string.

    Args:
        data (Dict): json Dict to parse.
        message_if_success (str, optional): Brief success-indicating message /
        keyword. Defaults to "Operation successful.".
        message_if_fail (str, optional): Brief failure-indicating message /
        keyword. Defaults to "Operation failed.".

    Returns:
        str: Brief success-or-failure-indicating sentence / keyword and
        detailed information in the next sentence if available.
    """
    potential_message = data.get("message")
    if data.get("status") == "Success":
        return compose_str(success_message, potential_message)
    return compose_str(
        fail_message, potential_message,
        fallback="No further details provided."
    )


def compose_str(
    main_str: str, potential_str: Optional[str], fallback: Optional[str] = None
) -> str:
    """Composes a string from main_str, potential_str and fallback.

    Args:
        main_str (str): the main string.
        potential_str (Optional[str]): data that may be a string or None.
        fallback (Optional[str], optional): a string to attach to main_str if
        potential_str is not a string. None if nothing should be attached.
        Defaults to None.

    Returns:
        str: main_str if potential_str and fallback are None. main_str
        followed by a space and potential_str if potential_str is a string.
        main_str followed by a space and fallback if potential_str is None and
        fallback is a string.
    """
    if isinstance(potential_str, str):
        return "%s %s" % (main_str, potential_str)
    if isinstance(fallback, str):
        return "%s %s" % (main_str, fallback)
    return main_str


def compose_json_reply(
    data: Dict,
    success_message: str = "Operation successful.",
    fail_message: str = "Operation failed.",
) -> Dict:
    """Enhances the json reply from misty's REST API with success_message
    in case of success and with fail_message otherwise.

    Args:
        data (Dict): json Dict to enhance.
        success_message (str, optional): A message / keyword to append in case
        of success. Defaults to "Operation successful.".
        fail_message (str, optional): A message / keyword to append in case
        of failure. Defaults to "Operation failed.".

    Returns:
        Dict: The enhanced json dictionary.
    """
    potential_message = data.get("message")
    status = data.get("status")
    if status == "Success":
        message = compose_str(success_message, potential_message)
    else:
        message = compose_str(
            fail_message, potential_message,
            fallback="No further details provided."
        )
    return {"status": status, "message": message}


def success_parser_message(message: Dict) -> Tuple[Dict, bool]:
    """Parses the successfulness of an action from the server's json reply.

    Args:
        message (Dict): the json dict to parse.

    Returns:
        Tuple[Dict, bool]: a tuple of a dictionary with keys "successful"
        (bool) and "message" (Dict) and a boolean indicating successfulness.
    """
    st = message.pop("status", None)
    if st == "Success":
        return {"successful": True, "message": message}, True
    return {"successful": False, "message": message}, False


def success_parser_from_dicts(**messages) -> Dict:
    """Parses the successfulness of a dictionary of actions, where the keyword
    is the action name and the value is the json reply from the server.
    Overall success is only true if all actions were successful.

    Returns:
        Dict: the dictionary of the keys "overall_success" (bool) and action
        names that contain the success-parsed message (a dict with keys
        "successful" (bool) and "message" (dict)).
    """
    status_dict = {}
    overall_success = True
    for name, message in messages.items():
        new_message, success = success_parser_message(message)
        status_dict[name] = new_message
        if not success:
            overall_success = False
    status_dict["overall_success"] = overall_success
    return status_dict


def success_parser_from_list(message_list: List[Dict]) -> Dict:
    """Parses the successfulness of a list of dictionaries of an action, where
    the keyword is the action name and the value is the json reply from the
    server. Overall success is only true if all actions were successful.

    Args:
        message_list (List[Dict]): the list of dictionaries of an action,
        where the keyword is the action name and the value is the json reply
        from the server.

    Returns:
        Dict: the dictionary of the keys "overall_success" (bool) and
        "actions" whose value is a list of dictionaries with names that
        contain the success-parsed message (a dict with keys "successful"
        (bool) and "message" (dict)).
    """
    status_dict = {"actions": []}
    overall_success = True
    for event in message_list:
        for name, message in event.items():
            new_message, success = success_parser_message(message)
        status_dict["actions"].append((name, new_message))
        if not success:
            overall_success = False
    status_dict["overall_success"] = overall_success
    return status_dict
